@@ -12,8 +12,13 @@ THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLI
 
 #include <vsgXchange/images.h>
 
+#include <vsg/core/Array.h>
+#include <vsg/core/MipmapLayout.h>
 #include <vsg/io/FileSystem.h>
 #include <vsg/io/stream.h>
+#include <vsg/state/ImageInfo.h>
+#include <vsg/utils/CommandLine.h>
+#include <vsg/utils/CoordinateSpace.h>
 
 #include <cstring>
 
@@ -46,6 +51,8 @@ namespace
         {tinyddsloader::DDSFile::DXGIFormat::R8G8_SInt, VK_FORMAT_R8G8_SINT},
         {tinyddsloader::DDSFile::DXGIFormat::BC7_UNorm, VK_FORMAT_BC7_UNORM_BLOCK},
         {tinyddsloader::DDSFile::DXGIFormat::BC7_UNorm_SRGB, VK_FORMAT_BC7_SRGB_BLOCK},
+        {tinyddsloader::DDSFile::DXGIFormat::BC6H_UF16, VK_FORMAT_BC6H_UFLOAT_BLOCK},
+        {tinyddsloader::DDSFile::DXGIFormat::BC6H_SF16, VK_FORMAT_BC6H_SFLOAT_BLOCK},
         {tinyddsloader::DDSFile::DXGIFormat::BC5_UNorm, VK_FORMAT_BC5_UNORM_BLOCK},
         {tinyddsloader::DDSFile::DXGIFormat::BC5_SNorm, VK_FORMAT_BC5_SNORM_BLOCK},
         {tinyddsloader::DDSFile::DXGIFormat::BC4_UNorm, VK_FORMAT_BC4_UNORM_BLOCK},
@@ -59,10 +66,11 @@ namespace
         {tinyddsloader::DDSFile::DXGIFormat::R16G16B16A16_Float, VK_FORMAT_R16G16B16A16_SFLOAT},
         {tinyddsloader::DDSFile::DXGIFormat::R32G32B32A32_Float, VK_FORMAT_R32G32B32A32_SFLOAT}};
 
-    uint8_t* allocateAndCopyToContiguousBlock(tinyddsloader::DDSFile& ddsFile)
+    std::pair<uint8_t*, vsg::ref_ptr<vsg::MipmapLayout>> allocateAndCopyToContiguousBlock(tinyddsloader::DDSFile& ddsFile, const vsg::Data::Properties& layout)
     {
-        const auto numMipMaps = ddsFile.GetMipCount();
+        const auto numMipMaps = layout.mipLevels;
         const auto numArrays = ddsFile.GetArraySize();
+
         size_t totalSize = 0;
         for (uint32_t i = 0; i < numMipMaps; ++i)
         {
@@ -73,10 +81,14 @@ namespace
             }
         }
 
-        if (totalSize == 0) return nullptr;
+        if (totalSize == 0) return {nullptr, nullptr};
 
         auto raw = new uint8_t[totalSize];
 
+        auto mipmapLayout = vsg::MipmapLayout::create(numMipMaps);
+        auto mipmapItr = mipmapLayout->begin();
+
+        uint32_t offset = 0;
         uint8_t* image_ptr = raw;
         for (uint32_t i = 0; i < numMipMaps; ++i)
         {
@@ -84,12 +96,19 @@ namespace
             {
                 const auto data = ddsFile.GetImageData(i, j);
 
+                if (j == 0)
+                {
+                    (*mipmapItr++).set(data->m_width, data->m_height, data->m_depth, offset);
+                }
+
                 std::memcpy(image_ptr, data->m_mem, data->m_memSlicePitch);
 
+                offset += static_cast<uint32_t>(data->m_memSlicePitch);
                 image_ptr += data->m_memSlicePitch;
             }
         }
-        return raw;
+
+        return {raw, mipmapLayout};
     }
 
     int computeImageViewType(tinyddsloader::DDSFile& ddsFile)
@@ -112,24 +131,45 @@ namespace
         return -1;
     }
 
+    static void process_image_format(vsg::ref_ptr<const vsg::Options> options, VkFormat& format)
+    {
+        if (!options) return;
+
+        vsg::CoordinateSpace coordinateSpace;
+        if (options->getValue(vsgXchange::dds::image_format, coordinateSpace))
+        {
+            if (coordinateSpace == vsg::CoordinateSpace::sRGB)
+                format = vsg::uNorm_to_sRGB(format);
+            else if (coordinateSpace == vsg::CoordinateSpace::LINEAR)
+                format = vsg::sRGB_to_uNorm(format);
+        }
+    }
+
     vsg::ref_ptr<vsg::Data> readCompressed(tinyddsloader::DDSFile& ddsFile, VkFormat targetFormat)
     {
         const auto width = ddsFile.GetWidth();
         const auto height = ddsFile.GetHeight();
-        const auto numMipMaps = ddsFile.GetMipCount();
         const auto numArrays = ddsFile.GetArraySize();
 
-        auto raw = allocateAndCopyToContiguousBlock(ddsFile);
+        auto formatTraits = vsg::getFormatTraits(targetFormat);
 
-        vsg::ref_ptr<vsg::Data> vsg_data;
+        uint32_t widthInBlocks = (width + formatTraits.blockWidth - 1) / formatTraits.blockWidth;
+        uint32_t heightInBlocks = (height + formatTraits.blockHeight - 1) / formatTraits.blockHeight;
+
+        uint32_t numMipMaps = ddsFile.GetMipCount();
 
         vsg::Data::Properties layout;
         layout.format = targetFormat;
-        layout.maxNumMipmaps = numMipMaps;
-        layout.blockWidth = 4;
-        layout.blockHeight = 4;
+        layout.mipLevels = numMipMaps;
+        layout.blockWidth = formatTraits.blockWidth;
+        layout.blockHeight = formatTraits.blockHeight;
+        layout.blockDepth = formatTraits.blockDepth;
         layout.imageViewType = computeImageViewType(ddsFile);
 
+        auto [raw, mipmapLayout] = allocateAndCopyToContiguousBlock(ddsFile, layout);
+
+        vsg::ref_ptr<vsg::Data> vsg_data;
+
         switch (targetFormat)
         {
         case VK_FORMAT_BC1_RGBA_UNORM_BLOCK:
@@ -137,9 +177,9 @@ namespace
         case VK_FORMAT_BC4_SNORM_BLOCK:
         case VK_FORMAT_BC4_UNORM_BLOCK:
             if (numArrays > 1)
-                vsg_data = vsg::block64Array3D::create(width / layout.blockWidth, height / layout.blockHeight, numArrays, reinterpret_cast<vsg::block64*>(raw), layout);
+                vsg_data = vsg::block64Array3D::create(widthInBlocks, heightInBlocks, numArrays, reinterpret_cast<vsg::block64*>(raw), layout, mipmapLayout);
             else
-                vsg_data = vsg::block64Array2D::create(width / layout.blockWidth, height / layout.blockHeight, reinterpret_cast<vsg::block64*>(raw), layout);
+                vsg_data = vsg::block64Array2D::create(widthInBlocks, heightInBlocks, reinterpret_cast<vsg::block64*>(raw), layout, mipmapLayout);
             break;
         case VK_FORMAT_BC2_UNORM_BLOCK:
         case VK_FORMAT_BC2_SRGB_BLOCK:
@@ -147,22 +187,30 @@ namespace
         case VK_FORMAT_BC3_SRGB_BLOCK:
         case VK_FORMAT_BC5_UNORM_BLOCK:
         case VK_FORMAT_BC5_SNORM_BLOCK:
+        case VK_FORMAT_BC6H_UFLOAT_BLOCK:
+        case VK_FORMAT_BC6H_SFLOAT_BLOCK:
         case VK_FORMAT_BC7_UNORM_BLOCK:
         case VK_FORMAT_BC7_SRGB_BLOCK:
             if (numArrays > 1)
-                vsg_data = vsg::block128Array3D::create(width / layout.blockWidth, height / layout.blockHeight, numArrays, reinterpret_cast<vsg::block128*>(raw), layout);
+                vsg_data = vsg::block128Array3D::create(widthInBlocks, heightInBlocks, numArrays, reinterpret_cast<vsg::block128*>(raw), layout, mipmapLayout);
             else
-                vsg_data = vsg::block128Array2D::create(width / layout.blockWidth, height / layout.blockHeight, reinterpret_cast<vsg::block128*>(raw), layout);
+                vsg_data = vsg::block128Array2D::create(widthInBlocks, heightInBlocks, reinterpret_cast<vsg::block128*>(raw), layout, mipmapLayout);
             break;
         default:
             std::cerr << "dds::readCompressed() Format is not supported yet: " << (uint32_t)targetFormat << std::endl;
             break;
         }
+#if 0
+        if (vsg_data && mipmapLayout)
+        {
+            vsg_data->setObject("mipmapLayout", mipmapLayout);
+        }
+#endif
 
         return vsg_data;
     }
 
-    vsg::ref_ptr<vsg::Data> readDds(tinyddsloader::DDSFile& ddsFile)
+    vsg::ref_ptr<vsg::Data> readDds(tinyddsloader::DDSFile& ddsFile, vsg::ref_ptr<const vsg::Options> options)
     {
         const auto width = ddsFile.GetWidth();
         const auto height = ddsFile.GetHeight();
@@ -173,7 +221,6 @@ namespace
         const auto dim = ddsFile.GetTextureDimension();
         const auto numArrays = ddsFile.GetArraySize();
 
-        // const auto valueCount = vsg::Data::computeValueCountIncludingMipmaps(width, height, depth, numMipMaps) * numArrays;
         // const auto bpp = tinyddsloader::DDSFile::GetBitsPerPixel(format);
         //std::cerr << "Fileinfo width: " << width << ", height: " << height << ", depth: " << depth << ", count: " << valueCount
         //          << ", format: " << (int)format << ", bpp: " << tinyddsloader::DDSFile::GetBitsPerPixel(format)
@@ -182,35 +229,34 @@ namespace
 
         if (auto it = kFormatMap.find(format); it != kFormatMap.end())
         {
+            vsg::ref_ptr<vsg::Data> vsg_data;
             if (isCompressed)
             {
-                return readCompressed(ddsFile, it->second);
+                vsg_data = readCompressed(ddsFile, it->second);
             }
             else
             {
-                auto raw = allocateAndCopyToContiguousBlock(ddsFile);
-
-                vsg::ref_ptr<vsg::Data> vsg_data;
-
                 vsg::Data::Properties layout;
                 layout.format = it->second;
-                layout.maxNumMipmaps = numMipMaps;
+                layout.mipLevels = numMipMaps;
                 layout.imageViewType = computeImageViewType(ddsFile);
 
+                auto [raw, mipmapLayout] = allocateAndCopyToContiguousBlock(ddsFile, layout);
+
                 switch (dim)
                 {
                 case tinyddsloader::DDSFile::TextureDimension::Texture1D:
                     switch (layout.format)
                     {
-                        case VK_FORMAT_R32G32B32A32_SFLOAT:
-                            vsg_data = vsg::vec4Array::create(width, reinterpret_cast<vsg::vec4*>(raw), layout);
-                            break;
-                        case VK_FORMAT_R16G16B16A16_SFLOAT:
-                            vsg_data = vsg::usvec4Array::create(width, reinterpret_cast<vsg::usvec4*>(raw), layout);
-                            break;
-                        default:
-                            vsg_data = vsg::ubvec4Array::create(width, reinterpret_cast<vsg::ubvec4*>(raw), layout);
-                            break;
+                    case VK_FORMAT_R32G32B32A32_SFLOAT:
+                        vsg_data = vsg::vec4Array::create(width, reinterpret_cast<vsg::vec4*>(raw), layout, mipmapLayout);
+                        break;
+                    case VK_FORMAT_R16G16B16A16_SFLOAT:
+                        vsg_data = vsg::usvec4Array::create(width, reinterpret_cast<vsg::usvec4*>(raw), layout, mipmapLayout);
+                        break;
+                    default:
+                        vsg_data = vsg::ubvec4Array::create(width, reinterpret_cast<vsg::ubvec4*>(raw), layout, mipmapLayout);
+                        break;
                     }
                     break;
                 case tinyddsloader::DDSFile::TextureDimension::Texture2D:
@@ -218,56 +264,62 @@ namespace
                     {
                         switch (layout.format)
                         {
-                            case VK_FORMAT_R32G32B32A32_SFLOAT:
-                                vsg_data = vsg::vec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::vec4*>(raw), layout);
-                                break;
-                            case VK_FORMAT_R16G16B16A16_SFLOAT:
-                                vsg_data = vsg::usvec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::usvec4*>(raw), layout);
-                                break;
-                            default:
-                                vsg_data = vsg::ubvec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::ubvec4*>(raw), layout);
-                                break;
+                        case VK_FORMAT_R32G32B32A32_SFLOAT:
+                            vsg_data = vsg::vec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::vec4*>(raw), layout, mipmapLayout);
+                            break;
+                        case VK_FORMAT_R16G16B16A16_SFLOAT:
+                            vsg_data = vsg::usvec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::usvec4*>(raw), layout, mipmapLayout);
+                            break;
+                        default:
+                            vsg_data = vsg::ubvec4Array3D::create(width, height, numArrays, reinterpret_cast<vsg::ubvec4*>(raw), layout, mipmapLayout);
+                            break;
                         }
                     }
                     else
                     {
                         switch (layout.format)
                         {
-                            case VK_FORMAT_R32G32B32A32_SFLOAT:
-                                vsg_data = vsg::vec4Array2D::create(width, height, reinterpret_cast<vsg::vec4*>(raw), layout);
-                                break;
-                            case VK_FORMAT_R16G16B16A16_SFLOAT:
-                                vsg_data = vsg::usvec4Array2D::create(width, height, reinterpret_cast<vsg::usvec4*>(raw), layout);
-                                break;
-                            default:
-                                vsg_data = vsg::ubvec4Array2D::create(width, height, reinterpret_cast<vsg::ubvec4*>(raw), layout);
-                               break;
+                        case VK_FORMAT_R32G32B32A32_SFLOAT:
+                            vsg_data = vsg::vec4Array2D::create(width, height, reinterpret_cast<vsg::vec4*>(raw), layout, mipmapLayout);
+                            break;
+                        case VK_FORMAT_R16G16B16A16_SFLOAT:
+                            vsg_data = vsg::usvec4Array2D::create(width, height, reinterpret_cast<vsg::usvec4*>(raw), layout, mipmapLayout);
+                            break;
+                        default:
+                            vsg_data = vsg::ubvec4Array2D::create(width, height, reinterpret_cast<vsg::ubvec4*>(raw), layout, mipmapLayout);
+                            break;
                         }
                     }
                     break;
                 case tinyddsloader::DDSFile::TextureDimension::Texture3D:
                     switch (layout.format)
                     {
-                        case VK_FORMAT_R32G32B32A32_SFLOAT:
-                            vsg_data = vsg::vec4Array3D::create(width, height, depth, reinterpret_cast<vsg::vec4*>(raw), layout);
-                            break;
-                        case VK_FORMAT_R16G16B16A16_SFLOAT:
-                            vsg_data = vsg::usvec4Array3D::create(width, height, depth, reinterpret_cast<vsg::usvec4*>(raw), layout);
-                            break;
-                        default:
-                            vsg_data = vsg::ubvec4Array3D::create(width, height, depth, reinterpret_cast<vsg::ubvec4*>(raw), layout);
-                            break;
+                    case VK_FORMAT_R32G32B32A32_SFLOAT:
+                        vsg_data = vsg::vec4Array3D::create(width, height, depth, reinterpret_cast<vsg::vec4*>(raw), layout, mipmapLayout);
+                        break;
+                    case VK_FORMAT_R16G16B16A16_SFLOAT:
+                        vsg_data = vsg::usvec4Array3D::create(width, height, depth, reinterpret_cast<vsg::usvec4*>(raw), layout, mipmapLayout);
+                        break;
+                    default:
+                        vsg_data = vsg::ubvec4Array3D::create(width, height, depth, reinterpret_cast<vsg::ubvec4*>(raw), layout, mipmapLayout);
+                        break;
                     }
                     break;
                 case tinyddsloader::DDSFile::TextureDimension::Unknown:
                     std::cerr << "dds::readDds() Num of dimension (" << (uint32_t)dim << ")  not supported." << std::endl;
                     break;
                 }
-
-                //std::cout << "* Finish: " << valueCount * valueSize << ", " << vsg_data->dataSize() << std::endl;
-
-                return vsg_data;
+#if 0
+                if (vsg_data && mipmapLayout)
+                {
+                    vsg_data->setObject("mipmapLayout", mipmapLayout);
+                }
+#endif
             }
+
+            if (options && vsg_data) process_image_format(options, vsg_data->properties.format);
+
+            return vsg_data;
         }
         else
         {
@@ -276,6 +328,7 @@ namespace
 
         return {};
     }
+
 } // namespace
 
 using namespace vsgXchange;
@@ -297,7 +350,7 @@ vsg::ref_ptr<vsg::Object> dds::read(const vsg::Path& filename, vsg::ref_ptr<cons
     std::ifstream ifs(filenameToUse, std::ios_base::binary);
     if (const auto result = ddsFile.Load(ifs); result == tinyddsloader::Success)
     {
-        return readDds(ddsFile);
+        return readDds(ddsFile, options);
     }
     else
     {
@@ -322,7 +375,7 @@ vsg::ref_ptr<vsg::Object> dds::read(std::istream& fin, vsg::ref_ptr<const vsg::O
     tinyddsloader::DDSFile ddsFile;
     if (const auto result = ddsFile.Load(fin); result == tinyddsloader::Success)
     {
-        return readDds(ddsFile);
+        return readDds(ddsFile, options);
     }
     else
     {
@@ -347,7 +400,7 @@ vsg::ref_ptr<vsg::Object> dds::read(const uint8_t* ptr, size_t size, vsg::ref_pt
     tinyddsloader::DDSFile ddsFile;
     if (const auto result = ddsFile.Load(ptr, size); result == tinyddsloader::Success)
     {
-        return readDds(ddsFile);
+        return readDds(ddsFile, options);
     }
     else
     {
@@ -371,5 +424,14 @@ bool dds::getFeatures(Features& features) const
     {
         features.extensionFeatureMap[ext] = static_cast<vsg::ReaderWriter::FeatureMask>(vsg::ReaderWriter::READ_FILENAME | vsg::ReaderWriter::READ_ISTREAM | vsg::ReaderWriter::READ_MEMORY);
     }
+
+    features.optionNameTypeMap[dds::image_format] = vsg::type_name<vsg::CoordinateSpace>();
+
     return true;
 }
+
+bool dds::readOptions(vsg::Options& options, vsg::CommandLine& arguments) const
+{
+    bool result = arguments.readAndAssign<vsg::CoordinateSpace>(dds::image_format, &options);
+    return result;
+}
