@@ -10,12 +10,15 @@ THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLI
 
 </editor-fold> */
 
+#include <vsgXchange/gltf.h>
 #include <vsgXchange/models.h>
 
 using namespace vsgXchange;
 
 models::models()
 {
+    add(gltf::create());
+
 #ifdef vsgXchange_assimp
     add(assimp::create());
 #endif
